from functools import partial
from typing import List, Optional

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import ProcessState, RequiresNoFinalisation
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import set_object_label

from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.cli.options import TypedOption
from wai.common.geometry import Polygon, Point

from .._format import YOLOODFormat, YOLOObject


class FromYOLOOD(
    RequiresNoFinalisation,
    ProcessorComponent[YOLOODFormat, ImageObjectDetectionInstance]
):
    """
    Converter from YOLO annotations to internal format.
    """
    # Path to the labels file
    labels_file: Optional[str] = TypedOption(
        "-l", "--labels",
        type=str,
        metavar="PATH",
        help="Path to the labels file"
    )

    # Mapping from class index to label
    labels: List[str] = ProcessState(lambda self: self.read_labels_file())

    def process_element(
            self,
            element: YOLOODFormat,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        # Unpack the external format
        image_info, yolo_objects = element

        # Convert YOLO objects to located objects
        located_objects = None
        if len(yolo_objects) > 0:
            to_located_object = partial(self.to_located_object, image_width=image_info.width, image_height=image_info.height)
            located_objects = LocatedObjects(map(to_located_object, yolo_objects))

        then(
            ImageObjectDetectionInstance(
                image_info,
                located_objects
            )
        )

    def read_labels_file(self) -> List[str]:
        """
        Parses the labels file if one is given.

        :return:
                    The label mapping.
        """
        # If no label file is given, return an empty mapping
        if self.labels_file is None:
            return []

        # Read and parse the labels file
        with open(self.labels_file, "r") as labels_file:
            return [x.strip() for x in labels_file.read().split(",")]

    def to_located_object(self, object: YOLOObject, *, image_width: int, image_height: int) -> LocatedObject:
        """
        Converts the YOLO object to a located object.

        :param object:
                    The YOLO object.
        :return:
                    The located object.
        """
        # Get the object label (just uses the class index if no mapping is provided)
        label: str = self.labels[object.class_index] if len(self.labels) > 0 else str(object.class_index)

        # Get the boundary co-ordinates
        width = round(object.width * image_width)
        height = round(object.height * image_height)
        x_min = round(object.centre_x * image_width - width / 2)
        y_min = round(object.centre_y * image_height - height / 2)

        # Create the located object
        located_object = LocatedObject(x_min, y_min, width, height)
        set_object_label(located_object, label)

        # polygon information?
        if object.has_polygon():
            points = []
            for i in range(len(object.poly_x)):
                points.append(Point(round(object.poly_x[i] * image_width), round(object.poly_y[i] * image_height)))
            located_object.set_polygon(Polygon(*points))

        return located_object



"""
Conversion components to/from the YOLO object detection format.
"""
from ._FromYOLOOD import FromYOLOOD
from ._ToYOLOOD import ToYOLOOD
from ._YOLOODReader import YOLOODReader
from ._YOLOODWriter import YOLOODWriter



import os
from typing import List, Optional

from wai.annotations.core.component.util import AnnotationFileProcessor
from wai.annotations.core.stream import ThenFunction
from wai.annotations.domain.image import Image
from wai.annotations.domain.image.util import get_associated_image

from wai.common.cli.options import TypedOption

from .._format import YOLOODFormat, YOLOObject


class YOLOODReader(AnnotationFileProcessor[YOLOODFormat]):
    """
    Reader of YOLO object-detection files.
    """
    # The relative path to the data image files from the annotation images
    relative_path_to_data_images: Optional[str] = TypedOption(
        "--image-path-rel",
        type=str,
        metavar="PATH",
        help="Relative path to image files from annotations"
    )

    def read_annotation_file(
            self,
            filename: str,
            then: ThenFunction[YOLOODFormat]
    ):

        # Split the filename into path, basename, ext
        path, basename = os.path.split(filename)
        basename, extension = os.path.splitext(basename)

        # Use the provided relative path if given, otherwise default to the same
        # sub-structure from a "labels" directory down in an "images" directory
        relative_path = self.relative_path_to_data_images
        if relative_path is None:
            relative_path = os.path.join("..", "images")
            abs_path = os.path.abspath(path)
            path_parts = os.path.split(abs_path)
            relevant_path_parts = []
            while abs_path:
                abs_path, path_part = os.path.split(abs_path)
                if path_part == "labels":
                    break
                relevant_path_parts.append(path_part)
            if len(relevant_path_parts) == len(path_parts):
                raise Exception(f"No 'labels' directory found in path of {abs_path}")
            for path_part in reversed(relevant_path_parts):
                relative_path = os.path.join("..", relative_path, path_part)

        # Join the path with the relative path to the data-image
        image_path = os.path.join(path, relative_path)

        # Find the image
        image_filename = get_associated_image(
            os.path.join(image_path, basename)
        )

        if image_filename is None:
            self.logger.warning("Failed to locate image for: %s" % filename)
            return

        # Read the YOLO annotations
        objects = []
        with open(filename, "r") as file:
            for line in file.readlines():
                objects.append(YOLOObject.from_string(line))

        # Read the image
        image = Image.from_file(image_filename)

        then((image, tuple(objects)))

    def read_negative_file(
            self,
            filename: str,
            then: ThenFunction[YOLOODFormat]
    ):
        image_info = Image.from_file(filename)

        then((image_info, None))



from dataclasses import dataclass
from typing import Tuple

from wai.annotations.domain.image import Image


@dataclass
class YOLOObject:
    """
    Internal representation of a YOLO annotation.
    """
    class_index: int
    centre_x: float = None
    centre_y: float = None
    width: float = None
    height: float = None
    poly_x: list = None
    poly_y: list = None

    @classmethod
    def from_string(cls, string: str):
        """
        Attempts to parse the given string.

        :param string: the string to parse
        :type string: str
        :return: the YOLOObject or an exception if failed to parse
        :rtype: YOLOObject
        """
        parts = string.strip().split(" ")
        # bbox format: <index> <center_x> <center_y> <widht> <height>
        if len(parts) == 5:
            return cls(
                int(parts[0]),
                float(parts[1]),
                float(parts[2]),
                float(parts[3]),
                float(parts[4]),
                None,
                None,
            )
        # polygon format: <index> <x0> <y0> <x1> <y1> ...
        elif (len(parts) > 5) and (len(parts) % 2 == 1):
            px = []
            py = []
            for i in range(1, len(parts), 2):
                px.append(float(parts[i]))
                py.append(float(parts[i+1]))
            w = max(px) - min(px)
            h = max(py) - min(py)
            return cls(
                int(parts[0]),
                min(px) + w / 2,
                min(py) + h / 2,
                w,
                h,
                px,
                py,
            )
        else:
            raise Exception("Neither in bbox nor polygon format: %s" % string)

    def has_polygon(self):
        """
        Returns whether polygon information is present.

        :return: True if present
        :rtype: bool
        """
        return (self.poly_x is not None) and (self.poly_y is not None)

    def to_bbox(self):
        """
        Returns the bbox line representation.

        :return: the generated line representation
        :rtype: str
        """
        return f"{self.class_index} {self.centre_x} {self.centre_y} {self.width} {self.height}"

    def to_polygon(self):
        """
        Returns the polygon line representation.

        :return: the generated line representation
        :rtype: str
        """
        parts = ["%d" % self.class_index]
        if self.has_polygon():
            for i in range(len(self.poly_x)):
                parts.append("%f %f" % (self.poly_x[i], self.poly_y[i]))
        else:
            x = self.centre_x - self.width / 2
            y = self.centre_y - self.height / 2
            parts.append("%f %f" % (x, y))
            parts.append("%f %f" % (x + self.width, y))
            parts.append("%f %f" % (x + self.width, y + self.height))
            parts.append("%f %f" % (x, y + self.height))
        return " ".join(parts)

    def to_str(self, use_polygon):
        """
        Turns the data into a line representation.

        :param use_polygon: whether to use polygon format or bbox format
        :type use_polygon: bool
        :return: the generated line
        :rtype: str
        """
        if use_polygon:
            return self.to_polygon()
        else:
            return self.to_bbox()

    def __str__(self):
        """
        Returns the polygon representation if possible otherwise the bbox one.

        :return: the representation as singe line
        :rtype: str
        """
        if self.has_polygon():
            return self.to_polygon()
        else:
            return self.to_bbox()


YOLOODFormat = Tuple[Image, Tuple[YOLOObject]]



import os
from functools import partial

from wai.annotations.core.component.util import (
    SeparateFileWriter,
    SplitSink,
    SplitState,
    ExpectsDirectory,
    RequiresNoSplitFinalisation
)

from .._format import YOLOODFormat


class YOLOODWriter(
    ExpectsDirectory,
    RequiresNoSplitFinalisation,
    SeparateFileWriter[YOLOODFormat],
    SplitSink[YOLOODFormat]
):
    """
    Writer of YOLO files.
    """
    labels_split_path: str = SplitState(lambda self: self.split_path("labels"))
    images_split_path: str = SplitState(lambda self: self.split_path("images"))

    def consume_element_for_split(
            self,
            element: YOLOODFormat
    ):
        # Unpack the instance
        image_info, yolo_objects = element

        # Write the image
        self.write_data_file(image_info, self.images_split_path)

        # If the image is a negative, skip writing the annotations
        if len(yolo_objects) == 0:
            return

        # Format the filename
        labels_filename = f"{os.path.splitext(image_info.filename)[0]}.txt"

        # Write the annotations file
        with open(os.path.join(self.labels_split_path, labels_filename), "w") as labels_file:
            labels_file.write("\n".join(map(str, yolo_objects)))

    @classmethod
    def get_help_text_for_output_option(cls) -> str:
        return "output directory to write images and annotations to"

    def split_path(self, path: str) -> str:
        """
        Helper function which both formats and makes a split directory.

        :param path:
                    The sub-directory in which to make the split directories.
        :return:
                    The split directory.
        """
        split_base_path = os.path.join(self.output_path, path)
        self.create_split_directories(split_base_path)
        return self.get_split_path(self.split_label, split_base_path)



from functools import partial
from typing import Dict, Optional

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import OutputElementType, ThenFunction, DoneFunction
from wai.annotations.core.stream.util import ProcessState
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.object_detection.util import get_object_label

from wai.common.adams.imaging.locateobjects import LocatedObject
from wai.common.cli.options import TypedOption, FlagOption

from .._format import YOLOODFormat, YOLOObject


class ToYOLOOD(
    ProcessorComponent[ImageObjectDetectionInstance, YOLOODFormat]
):
    """
    Converter from internal format to YOLO annotations.
    """
    # Path to the labels file to write
    labels_file: Optional[str] = TypedOption(
        "-l", "--labels",
        type=str,
        metavar="PATH",
        help="Path to the labels file to write"
    )

    # Path to the labels CSV file to write
    labels_csv_file: Optional[str] = TypedOption(
        "-c", "--labels-csv",
        type=str,
        metavar="PATH",
        help="Path to the labels CSV file to write"
    )

    # whether to output polygon format rather than bbox one
    use_polygon_format: Optional[bool] = FlagOption(
        "-p", "--use-polygon-format",
        help="Outputs the annotations in polygon format rather than bbox one."
    )

    # Label-index mapping accumulator
    labels: Dict[str, int] = ProcessState(lambda self: {})

    def process_element(
            self,
            element: ImageObjectDetectionInstance,
            then: ThenFunction[YOLOODFormat],
            done: DoneFunction
    ):
        image_info, located_objects = element

        if located_objects is None or len(located_objects) == 0:
            return then((image_info, tuple()))

        to_yolo_object = partial(self.to_yolo_object, image_width=image_info.width, image_height=image_info.height)
        yolo_objects = tuple(map(to_yolo_object, located_objects))

        then((image_info, yolo_objects))

    def finish(self, then: ThenFunction[OutputElementType], done: DoneFunction):
        # Write the labels file
        if self.labels_file is not None:
            with open(self.labels_file, "w") as labels_file:
                labels_file.write(",".join(self.labels.keys()))

        # Write the labels CSV file
        if self.labels_csv_file is not None:
            with open(self.labels_csv_file, "w") as labels_csv_file:
                labels_csv_file.write("Index,Label")
                for label, index in self.labels.items():
                    labels_csv_file.write(f"\n{index},{label}")

        done()

    def to_yolo_object(self, located_object: LocatedObject, *, image_width: int, image_height: int) -> YOLOObject:
        """
        Converts a single located object into a YOLO object.

        :param located_object:
                    The located object to convert.
        :param image_width:
                    The image width (for normalisation).
        :param image_height:
                    The image height (for normalisation).
        :return:
                    The YOLO object.
        """
        # Update the label mapping
        label = get_object_label(located_object)
        if label not in self.labels:
            self.labels[label] = len(self.labels)
        class_index = self.labels[label]

        px = None
        py = None
        if self.use_polygon_format:
            if located_object.has_polygon():
                px = located_object.get_polygon_x()
                py = located_object.get_polygon_y()
            else:
                l = located_object
                px = [l.x, l.x + l.width - 1, l.x + l.width - 1, l.x]
                py = [l.y, l.y, l.y + l.height - 1, l.y + l.height - 1]
            px = [x / image_width for x in px]
            py = [y / image_height for y in py]

        return YOLOObject(
            class_index,
            (located_object.x + located_object.width / 2) / image_width,
            (located_object.y + located_object.height / 2) / image_height,
            located_object.width / image_width,
            located_object.height / image_height,
            px,
            py,
        )
